package CicleBBNR;

import Case.ICase;
import CaseBase.ICaseBase;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import RetrieveReuse.*;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * Classe principal de l'algorisme BBNR
 */
public class BBNR implements IBBNR{
    /**
     * Classe per executar el BBNR
     * @param caseBase
     * @param dataset
     * @param fileTrain
     * @return 
     */
    @Override
    public Collection<ICase> executeBBNR(ICaseBase caseBase, String dataset, String fileTrain){		
        double[] minAttributeValues = caseBase.getMinAttributeValues();
        double[] maxAttributeValues = caseBase.getMaxAttributeValues();
        Retrieve retrieve = new Retrieve();
        Reuse reuse = new Reuse();
        int numK = 3;
        CoverageLiability coverageLiability = new CoverageLiability();
        
        /*
        * Blame-based Noise Reduction (BBNR) Algorithm
        * T, Training Set
        * For each c in T
        * CSet(c) = Coverage Set of c
        * LSet(c) = Liability Set of c
        * End-For
        *      
        * TSet = T sorted in descending order of LSet(c) size
        * c = first case in TSet
        *      
        * If |LSet(c)| >0
        *              TSet = TSet - {c}
        *              misClassifiedFlag = false
        *              For each x in CSet(c)
        *                      If x cannot be correctly classified by TSet
        *                              misClassifiedFlag = true
        *                              break
        *                      End-If
        *              End-For
        *              If misClassifiedFlag = true
        *                      TSet = TSet + {c}
        *              End-If
        *              c = next case in TSet
        * End-While
        * 
        * Return TSet
        */
        
        coverageLiability.createCoverageAndLiabilitySets(numK, caseBase, reuse, retrieve, minAttributeValues, maxAttributeValues);
        
        Map<String,Integer> casosSizeCoverage = new HashMap<String,Integer>();
        Map<String,Integer> casosSizeLiability = new HashMap<String,Integer>();
        
        for(ICase c: caseBase.trainCases()){
            casosSizeCoverage.put(c.getName(), c.getCoverageSet().size());
            casosSizeLiability.put(c.getName(), c.getLiabilitySet().size());
        }
        
        casosSizeCoverage = this.sortHashMapByValues(casosSizeCoverage);
        casosSizeLiability = this.sortHashMapByValues(casosSizeLiability);

        
        //Crear el conjunt de casos ordenats per size de Coverage
        Collection <ICase> newTrainingSet = new ArrayList<>();
        List<String> mapKeys = new ArrayList<>(casosSizeLiability.keySet());
        this.sortedByLiabilitySize(newTrainingSet, mapKeys, caseBase);
        
        //copiem el trainingset per poder esborrar i afegir casos sense problemes
        Collection<ICase> copySortSet = new ArrayList<>();
        Collection<ICase> removeCases = new ArrayList<>();
        for(ICase c: newTrainingSet){
            copySortSet.add(c);
        }

        for(ICase c: newTrainingSet){
            if(c.getLiabilitySet().size()>0){
                copySortSet.remove(c);
                removeCases.add(c);
                boolean missClassifiedFlag = false;
                for(ICase cov: c.getCoverageSet()){
                    ArrayList<ICase> kNeighbors = retrieve.retrieveCases(cov, copySortSet, minAttributeValues, maxAttributeValues, 1);
                    String classe;
                    classe = reuse.reviseCases(kNeighbors);
                    if(!classe.equals(cov.getAttributeClass())){
                        missClassifiedFlag = true;
                        break;
                    }
                }
                if(missClassifiedFlag){
                    copySortSet.add(c);
                    removeCases.remove(c);
                }
            }else{
                break;
            }
        }
                
        return removeCases;
    }		
    
    /**
     * Ordenar el conjunt de Liability
     * @param newTrainingSet
     * @param mapKeys
     * @param caseBase 
     */
    @Override
    public void sortedByLiabilitySize(Collection<ICase> newTrainingSet, List<String> mapKeys, ICaseBase caseBase){
        
        
        //Creem el conjunt de Casos ordenats
        for(int i=mapKeys.size()-1;i>=0;i--){
            String searchCase = mapKeys.get(i);
            
            boolean trobat = false;
            
            Iterator itr = caseBase.trainCases().iterator();
            while(itr.hasNext() && !trobat) {
               ICase element = (ICase) itr.next();
               if(element.getName().equals(searchCase)){
                    newTrainingSet.add(element);
                    trobat = true;
                }
            }
        }
        
    }
    
    /**
     * Ordenar un Map
     * @param passedMap
     * @return 
     */
    @Override
    public LinkedHashMap<String, Integer> sortHashMapByValues(Map<String, Integer> passedMap) {
        List<String> mapKeys = new ArrayList<>(passedMap.keySet());
        List<Integer> mapValues = new ArrayList<>(passedMap.values());
        Collections.sort(mapValues);
        Collections.sort(mapKeys);

        LinkedHashMap<String, Integer> sortedMap = new LinkedHashMap<>();

        Iterator<Integer> valueIt = mapValues.iterator();
        while (valueIt.hasNext()) {
            Integer val = valueIt.next();
            Iterator<String> keyIt = mapKeys.iterator();

            while (keyIt.hasNext()) {
                String key = keyIt.next();
                Integer comp1 = passedMap.get(key);
                Integer comp2 = val;

                if (comp1.equals(comp2)) {
                    keyIt.remove();
                    sortedMap.put(key, val);
                    break;
                }
            }
        }
        return sortedMap;
    }
}

